use snafu::ResultExt;
use qbittorrent_web_api::Api;
use qbittorrent_web_api::api_impl::Error as QBittorrentError;
use tokio::runtime::Builder;

use crate::config::Config;
use crate::error::{Error, ApiSnafu as IOError, InternalApiSnafu as ApiError};

pub fn blocking_runtime() -> std::io::Result<tokio::runtime::Runtime> {
    Builder::new_current_thread()
        .enable_all()
        .build()
}

#[derive(Debug)]
pub struct UnauthenticatedRawApiClient {
    pub rt: tokio::runtime::Runtime,
}

impl UnauthenticatedRawApiClient {
    /// Initialize a blocking runtime for the API client
    pub fn new() -> Result<UnauthenticatedRawApiClient, std::io::Error> {
        Ok(UnauthenticatedRawApiClient {
            rt: blocking_runtime()?,
        })
    }

    /// Login into a qBittorrent backend and return a proper ApiClient instance
    pub fn login(self, host: &str, login: &str, password: &str) -> Result<RawApiClient, QBittorrentError> {
        let api = self.rt.block_on(Api::login(host, login, password))?;
        Ok(RawApiClient {
            rt: self.rt,
            api,
        })
    }
}

#[derive(Debug)]
pub struct RawApiClient {
    rt: tokio::runtime::Runtime,
    api: qbittorrent_web_api::api_impl::Authenticated,
}

impl RawApiClient {
    pub fn add(&self, magnet: &str, paused: bool) -> Result<String, QBittorrentError> {
        let base_call = self.api.torrent_management();
        let call = if paused {
            base_call.add(magnet).paused("true")
        } else {
            base_call.add(magnet)
        };
        self.rt.block_on(call.send_raw())
    }

    pub fn get(&self, hash: &str) -> Result<String, QBittorrentError> {
        self.rt.block_on(self.api.torrent_management().properties_raw(hash))
    }

    pub fn list(&self) -> Result<String, QBittorrentError> {
        self.rt.block_on(self.api.torrent_management().info().send_raw())
    }

}

#[derive(Debug)]
/// ApiClient is a convenience struct around qbittorrent_web_api for use in async programs, using qbt::Error error types.
pub struct ApiClient {
    pub raw_api: RawApiClient,
}

impl ApiClient {
    /// Login to a qBittorrent backend and return a handle to the ApiClient
    pub fn new(host: &str, login: &str, password: &str) -> Result<ApiClient, Error> {
        let unauthenticated = UnauthenticatedRawApiClient::new().context(IOError)?;
        let authenticated = unauthenticated.login(host, login, password).map_err(|e| {
            match e {
                QBittorrentError::HttpError(_) => {
                    Error::FailedToReachAPI { source: e }
                }, QBittorrentError::InvalidUsernameOrPassword => {
                    Error::FailedLogin { user: login.to_string() }
                } _ => {
                    panic!("Cookie error");
                }
            }
        })?;
        Ok(ApiClient {
            raw_api: authenticated,
        })
    }

    pub fn from_config(config: &Config) -> Result<ApiClient, Error> {
        Self::new(&config.format_host(), &config.qbittorrent.login, &config.qbittorrent.password)
    }


    pub fn add(&self, magnet: &str, paused: bool) -> Result<(), Error> {
        let res = self.raw_api.add(magnet, paused).context(ApiError)?;
        if res == "Ok." {
            Ok(())
        } else {
            Err(Error::message(res.clone()))
        }
    }

    pub fn get(&self, hash: &str) -> Result<Option<String>, Error> {
        let res = self.raw_api.get(hash).context(ApiError)?;
        if res == "" {
            Ok(None)
        } else {
            Ok(Some(res))
        }
    }

    pub fn list(&self) -> Result<String, Error> {
        self.raw_api.list().context(ApiError)
    }

}

// TODO: typestate builder https://www.greyblake.com/blog/builder-with-typestate-in-rust/
//struct ApiClientBuilder {}

#[cfg(test)]
mod tests {
    use crate::utils::*;
    use crate::api::{ApiClient, UnauthenticatedRawApiClient, QBittorrentError};
    use crate::error::Error;

    #[test]
    fn raw_wrong_server() {
        let port = find_free_port();
        let api = UnauthenticatedRawApiClient::new().expect("IOERROR");
        let login = api.login(&format!("http://localhost:{}", &port), "admin", "adminadmin");
        assert!(login.is_err());
        let err = login.unwrap_err();
        match &err {
            QBittorrentError::HttpError(_) => {
                return;
            }, _ => {
                println!("{:?}", err);
                panic!("API CHANGE!");
            }
        }
    }

    #[test]
    fn wrong_server() {
        let port = find_free_port();
        let api = ApiClient::new(&format!("http://localhost:{}", &port), "admin", "adminadmin");
        assert!(api.is_err());
        let err = api.unwrap_err();
        match &err {
            Error::FailedToReachAPI { source: _ }  => {
                return;
            }, _ => {
                println!("{:?}", err);
                panic!("ERROR CONVERSION PROBLEM");
            }
        }
    }
}
